from typing import Final


__author__: Final = "Tsuyoshi Hombashi"
__copyright__: Final = f"Copyright 2020-2025, {__author__}"
__license__: Final = "MIT License"
__version__ = "0.6.3"
__maintainer__: Final = __author__
__email__: Final = "tsuyoshi.hombashi@gmail.com"
